#include <cmath>
#include <iostream>

#include <vector>

#include "mex.h"

#include "densecrf.h"
#include "fastmath.h"
#include "permutohedral.h"
#include "util.h"

// #define DEBUG_OUTPUT 
// Compile using    mex perform_dense_message_passing_mex_prealloc.cpp  permutohedral.cpp bipartitedensecrf.cpp densecrf.cpp util.cpp


/*
 * First parameter should be the model we are working on.
 */
#define MODELSET_ID    prhs[0]

/*
 * ACTION can be one of the following:
 *      'a'     ('alloc') for allocating data. NUM_LATTICES then has to be T
 *      'p'     ('process') to use pre-allocated lattices. Index with the number in NUM_LATTICES.
 *      'd'     ('dealloc') to deallocate all latices.
 */
#define ACTION      prhs[1]

// If mode = 'alloc', FEATURES_LATTICE_ID contains a cell of features.
// If mode = 'process', it contains the ID of the model to evaluate.
#define FEATURES_LATTICE_ID    prhs[2]

// If mode = 'alloc', VALS_IN_PARAMS contain the parameters (i.e., the weight)
// If mode = 'process', it contains the input values.
#define VALS_IN_PARAMS  prhs[3]

#define VALS_OUT    prhs[4] 

#define ACTION_ALLOCATE 'a'
#define ACTION_PROCESS 'p'
#define ACTION_DEALLOCATE 'd'


//#define DEBUG_OUTPUT 1

// One model consists of these variables.
typedef struct
{
    PottsPotential *ising_potentials;
    int num_lattices;
    float *F;
    float *in;
    float *out;
    float *tmp;
} Model;

std::vector<Model> all_models;


void init_model(int i)
{
    #ifdef DEBUG_OUTPUT
        mexPrintf("\tDEBUG:: Initializing model %d\n",i);
    #endif
    // Reszie all_models, to hold at least i elements
    if (i >= all_models.size())
    {
        Model m;
        m.ising_potentials = 0;
        m.num_lattices = 0;
        m.F = 0;
        m.in = 0;
        m.out = 0;
        m.tmp = 0;
        
        all_models.resize(i,m);
    }
    // else do nothing, we are already initialized.
    #ifdef DEBUG_OUTPUT
        mexPrintf("\tDEBUG:: Done initializing model %d\n",i);
    #endif    
}

void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{
    #ifdef DEBUG_OUTPUT
        mexPrintf("\tDEBUG:: Starting... \n");
    #endif    

    if (!mxIsNumeric(MODELSET_ID))
    {
        mexPrintf("ERROR:: Please give a numeric modelset ID as the first parameter.\n");
        return;
    }
    unsigned int current_modelset = (unsigned int) mxGetScalar(MODELSET_ID);
    
    #ifdef DEBUG_OUTPUT
        mexPrintf("\tDEBUG:: Modelset id = %d \n", current_modelset);
    #endif    
    
    // If the modelset is already initialized, do nothing.
    init_model(current_modelset);
    
    #ifdef DEBUG_OUTPUT
        mexPrintf("\tDEBUG:: Model init complete...\n", current_modelset);
    #endif    

    // Set up the references to the currently selected model, so we don't
    // have to bother with the modelset later.
    PottsPotential* &ising_potentials = all_models[current_modelset-1].ising_potentials;
    int &num_lattices = all_models[current_modelset-1].num_lattices;
    float* &F = all_models[current_modelset-1].F;
    float* &in = all_models[current_modelset-1].in;
    float* &out = all_models[current_modelset-1].out;
    float* &tmp = all_models[current_modelset-1].tmp;
    
    //double action = mxGetScalar(ACTION);
    char action = ((char*)mxGetChars(ACTION))[0];
    
    if(action == ACTION_ALLOCATE)
    {
        /*
         * Create the Ising models to reuse them later.
         * Also, allocate the utility arrays.
         */

        #ifdef DEBUG_OUTPUT
            mexPrintf("\t ============ Starting allocation. =============== \n");
        #endif
         
        // Make sure FEATURE_LATTICE_ID is a cell array
        if (!mxIsCell(FEATURES_LATTICE_ID))
        {
            mexPrintf("ERROR:: Argument is not a cell array.\n");
            return;
        }
        if (nrhs < 2)
        {
            mexPrintf("ERROR:: Too few parameters.\n");
            return;
        }
        
        num_lattices = mxGetNumberOfElements(FEATURES_LATTICE_ID);
        
        const int *param_dims = mxGetDimensions(VALS_IN_PARAMS);
        if (param_dims[0]*param_dims[1] != num_lattices)
        {
            mexPrintf("ERROR:: Number of feature arrays and number of parameters have to agree.\n");
            return;
        }
        
        double *parameter_array = (double*)mxGetData(VALS_IN_PARAMS);
        
        float weight = 1;
        
        // Determine weight, height (assumed to be constant), and maximum
        // number of features.
        // This is used to pre-allocate the utility arrays.
        int height=0, width=0, nFeaturesMax=0;
        int cur_height, cur_width, cur_nFeatures;
        
        for (int l=0; l < num_lattices; l++)
        {
            mxArray *feature_array = mxGetCell(FEATURES_LATTICE_ID,l);

            int number_of_dims	 = mxGetNumberOfDimensions(feature_array);
            const int *dim_array = mxGetDimensions(feature_array);

            cur_height		= dim_array[0];
            cur_width		= dim_array[1];
            if (number_of_dims > 2)
                cur_nFeatures	= dim_array[2];
                
            if (height > 0 || width > 0 || nFeaturesMax > 0)
            {
                if (height != cur_height || width != cur_width)
                {
                    mexPrintf("ERROR:: Different x/y dimensions in input arrays.\n");
                    return;
                }
                
                nFeaturesMax = cur_nFeatures > nFeaturesMax ? cur_nFeatures : nFeaturesMax;
            }
            else // Init
            {
                height = cur_height;
                width = cur_width;
                nFeaturesMax = cur_nFeatures;
            }
            
        }
        
        // Some static initialization
            
        // This *should* be 2.
        int value_size  = 2;

        #ifdef DEBUG_OUTPUT
            mexPrintf("%d %d %d %d \n", height, width, nFeaturesMax, value_size);
        #endif
        
        const int nPixels		= height*width;
        const int nFeaturePixels= nPixels*nFeaturesMax;
        const int nValuePixels	= nPixels*value_size;

        #ifdef DEBUG_OUTPUT
            mexPrintf("pixel numbers %d %d %d \n", nPixels, nFeaturePixels, nValuePixels);
        #endif
        
        
        // Allocate arrays with unchanged size
        F	= allocate(nFeaturePixels);
        in	= allocate(nValuePixels);
        out  = allocate(nValuePixels);	
        tmp = allocate(nValuePixels);

        #ifdef DEBUG_OUTPUT
            mexPrintf("allocate memory OK\n");
        #endif
        
        
        ising_potentials = new PottsPotential[num_lattices];
        
        
        for(int l=0; l < num_lattices; l++)
        {

            // Get weight parameter
            weight = parameter_array[l];
            
            #ifdef DEBUG_OUTPUT
                mexPrintf("get weight parameters OK %f, \n", weight);
            #endif	      
            
            // Get features for current model
            mxArray *feature_array = mxGetCell(FEATURES_LATTICE_ID,l);
            
            double *features;
            features  = (double*) mxGetPr(feature_array);

            #ifdef DEBUG_OUTPUT
                mexPrintf("get data pointers OK\n");
            #endif
            
            const int *dim_array = mxGetDimensions(feature_array);
            const int nFeatures = dim_array[2];

            
            // convert data to permutoheadra format
            for (int k=0; k < nFeatures; k++)
            {				
                for (int i=0; i<width; i++)
                {
                    for (int j=0; j<height; j++)
                    {
                        int k1 = k*nPixels + i*height +j;  //matlab
                        int k2 = (j*width+i)*nFeatures + k;		// permuhedra

                        #ifdef DEBUG_OUTPUT
                            if (k1 < 0 || k1 >= nFeaturePixels)
                                mexPrintf("out of range matlab index %d\n", k1);
                            if (k2 < 0 || k2 >= nFeaturePixels)
                                mexPrintf("out of range C++ index %d\n", k1);
                        #endif

                        F[k2] = (float) features[k1];
                    }
                }
            }

            #ifdef DEBUG_OUTPUT
                mexPrintf("get data OK\n");
            #endif
            
            ising_potentials[l].init(F, nFeatures, height*width, weight, true);
            
            #ifdef DEBUG_OUTPUT	
                mexPrintf("create Ising OK\n");
            #endif
            
        }
        #ifdef DEBUG_OUTPUT
            mexPrintf("\t ============ Finished allocation. =============== \n");
        #endif

    } // (action == ACTION_ALLOCATE)
    
    else if(action == ACTION_PROCESS)
    {
        /*
         * Process the lattice with ID given in LATTICE_INDEX.
         */
         
        #ifdef DEBUG_OUTPUT
            mexPrintf("\t ============ Starting process. =============== \n");
        #endif

        if (nrhs < 4)
        {
            mexPrintf("ERROR:: Too few parameters to process.\n");
            return;
        }
        
        if (!mxIsNumeric(FEATURES_LATTICE_ID))
        {
            mexPrintf("ERROR:: Argument is not a valid numeric index.\n");
            return;
        }
         
        unsigned int lattice_id = (unsigned int) mxGetScalar(FEATURES_LATTICE_ID);
         
        if (lattice_id > num_lattices)
        {
            mexPrintf("ERROR:: Not enough lattices were initialized.\n");
            return;
        }
        
        
        int height, width, value_size;
        int number_of_dims	 = mxGetNumberOfDimensions(VALS_IN_PARAMS);
        const int *dim_array = mxGetDimensions(VALS_IN_PARAMS);

        height		= dim_array[0];
        width		= dim_array[1];
        if (number_of_dims > 2)
            value_size	= dim_array[2];        
        
        int nPixels = width*height;
        
        double *vals_in = (double*) mxGetPr(VALS_IN_PARAMS);
        double *vals_out = (double*) mxGetPr(VALS_OUT);
                
        
        #ifdef DEBUG_OUTPUT
            mexPrintf("Starting to read data... ");
        #endif
        
        for (int k=0; k < value_size; k++)
        {
            for (int i=0; i<width; i++)
            {
                for (int j=0; j<height; j++)
                {
                    int k1 = k*nPixels+ i*height+j;	// matlab matrix index
                    int k2 = (j*width+i)*value_size + k;	// permutohedra.h array index
                    in[k2] = (float) vals_in[k1];
                }
            }
        }
            
        
        #ifdef DEBUG_OUTPUT
            mexPrintf("OK\n");
        #endif
        
        memset(out, 0, value_size*width*height*sizeof(float));
        memset(tmp, 0, value_size*width*height*sizeof(float));

		ising_potentials[lattice_id-1].apply(out, in, tmp, value_size);


        #ifdef DEBUG_OUTPUT
            mexPrintf("apply filtering OK\n");
        #endif

        for (int k=0; k < value_size; k++)
        {
            for (int i=0; i<width; i++)
            {
                for (int j=0; j<height; j++)
                {
					int k1 = k*nPixels+i*height+j;		// matlab matrix index
					int k2 = (j*width+i)*value_size +k;		// permutohedra.h array index
                    
                    /*
                    #ifdef DEBUG_OUTPUT
                        if (k1 < 0 || k1 >= nValuePixels)
                            printf("out of range matlab index %d\n", k1);
                        if (k2 < 0 || k2 >= nValuePixels)
                            printf("out of range C++ index %d\n", k1);
                    #endif
                    */
					
                    vals_out[k1] = (double) out[k2];

				}
            }
        }

        #ifdef DEBUG_OUTPUT
            mexPrintf("\t ============ Finished process. =============== \n");
        #endif


    } // (action == ACTION_PROCESS)

    else if(action == ACTION_DEALLOCATE)
    {
        #ifdef DEBUG_OUTPUT
            mexPrintf("\t ============ Starting deallocation. =============== \n");
        #endif
        
        if (F)
            deallocate(F);
            F = 0;
        if (in)
            deallocate(in);
            in = 0;
        if (out)
            deallocate(out);
            out = 0;
        if (tmp)
            deallocate(tmp);
            tmp = 0;
        
        if (ising_potentials)
            delete[] ising_potentials;
            ising_potentials = 0;

        #ifdef DEBUG_OUTPUT
            mexPrintf("\t ============ Finished deallocation. =============== \n");
        #endif

    }

    #ifdef DEBUG_OUTPUT
        printf("return OK\n"); 
    #endif


    return;
}

// It is important to note that the data inside the array is in column major order. 
// Instead of reading a matrix's values across and then down, the values are read down and then across. 
// This is contrary to how C indexing works and means that special care must be taken 
// when accessing the array's elements. To access the data inside of mxArrays, use the API functions (see below)
